@@ -240,6 +240,7 @@ MakeTemp(buf, complaint)
 char	*buf;
 const char	*complaint;
 {
+	const char *errfmt = "cannot create \"%s\" for %s: %s (%d)";
 #ifdef NO_MKSTEMP
 # ifdef NO_MKTEMP
 	/* Some systems don't provide mkstemp or mktemp (nor should
@@ -256,6 +257,7 @@ const char	*complaint;
 	for (;;) {
 		struct stat	sb;
 
+		jdbg("stat temp \"%s\"\n", buf);
 		if (stat(buf, &sb) < 0
 # ifdef ZTCDOS
 		/* Zortech yields ENOTDIR when path isn't found,
@@ -272,6 +274,7 @@ const char	*complaint;
 # else /* !O_EXCL */
 			int fd = creat(buf, 0600);
 # endif /* !O_EXCL */
+			jdbg("opened temp \"%s\" %d %d\n", buf, fd, errno);
 
 			if (fd != -1)
 				return fd;
@@ -279,7 +282,7 @@ const char	*complaint;
 		for (p = seq; ; ) {
 			if (*p == '\0') {
 				/* we ran out of possible names! */
-				complain(complaint, buf);
+				complain(errfmt, buf, complaint, "ran out of names", 0);
 				/*NOTREACHED*/
 			} else if (*p == '9') {
 				*p++ = '0';
@@ -292,6 +295,7 @@ const char	*complaint;
 # else /* !NO_MKTEMP */
 	int fd;
 
+	jdbg("mktemp \"%s\"\n", buf);
 	if (mktemp(buf) == NULL
 	|| -1 == (fd =
 #  ifdef O_EXCL
@@ -301,7 +305,7 @@ const char	*complaint;
 #  endif /* !O_EXCL */
 	))
 	{
-		complain(complaint, buf);
+		complain(errfmt, buf, complaint, strerror(errno), errno);
 		/* NOTREACHED */
 	}
 	return fd;
@@ -323,10 +327,11 @@ const char	*complaint;
 	int fd = mkstemp(buf);
 	int saved_errno = errno;
 
+	jdbg("mkstemp \"%s\" %d %d\n", buf, fd, saved_errno);
 	(void) umask(saved_umask);
 	errno = saved_errno;
 	if (fd == -1) {
-		complain(complaint, buf);
+		complain(errfmt, buf, complaint, strerror(errno), errno);
 		/*NOTREACHED*/
 	}
 	return fd;
@@ -1368,7 +1373,7 @@ tmpinit()
 #endif
 		);
 	tfname = copystr(buf);
-	tmpfd = MakeTemp(tfname, "cannot create tmp file \"%s\"");
+	tmpfd = MakeTemp(tfname, "editing");
 #ifdef RECOVER
 	rectmpname(strrchr(tfname, '/') + 1);
 #endif
@@ -1570,15 +1575,15 @@ private Block
 private daddr	next_bno = 0;
 
 /* Needed to comfort MS Visual C */
-private void blkio proto((Block *, SSIZE_T (*) ptrproto((int, UnivPtr, size_t))));
+private void blkio proto((Block *, JSSIZE_T (*) ptrproto((int, UnivPtr, size_t))));
 
 private void
 blkio(b, iofcn)
 register Block	*b;
-register SSIZE_T	(*iofcn) ptrproto((int, UnivPtr, size_t));
+register JSSIZE_T	(*iofcn) ptrproto((int, UnivPtr, size_t));
 {
 	off_t boff = bno_to_seek_off(b->b_bno);
-	SSIZE_T nb;
+	JSSIZE_T nb;
 	static bool first_time = YES;
 
 	if (first_time) {
@@ -1646,14 +1651,14 @@ SyncTmp()
 	 */
 	for (bno = 0; bno < next_bno; bno++) {
 		if ((b = lookup_block(bno)) != NULL && b->b_dirty) {
-			blkio(b, (SSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
+			blkio(b, (JSSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
 			b->b_dirty = NO;
 		}
 	}
 #else /* !MSDOS */
 	for (b = f_block; b != NULL; b = b->b_LRUnext)
 		if (b->b_dirty) {
-			blkio(b, (SSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
+			blkio(b, (JSSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
 			b->b_dirty = NO;
 		}
 #endif /* !MSDOS */
@@ -1718,7 +1723,7 @@ register Block	*bp;
 		bht[B_HASH(bp->b_bno)] = hp->b_HASHnext;
 
 	if (bp->b_dirty) {	/* do, now, the delayed write */
-		blkio(bp, (SSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
+		blkio(bp, (JSSIZE_T (*) ptrproto((int, UnivPtr, size_t)))write);
 		bp->b_dirty = NO;
 	}
 
@@ -1859,7 +1864,7 @@ file_backup(fname)
 char *fname;
 {
 # ifndef MSFILESYSTEM
-	SSIZE_T	rr;
+	JSSIZE_T	rr;
 	int
 		ffd,
 		bffd = 0;	/* avoid uninitialized complaint from gcc -W */
@@ -1909,7 +1914,7 @@ char *fname;
 		char	*p = buf;
 
 		while (rr > 0) {
-			SSIZE_T	wr = write(bffd, (UnivPtr) p, (size_t) rr);
+			JSSIZE_T	wr = write(bffd, (UnivPtr) p, (size_t) rr);
 
 			if (wr < 0) {
 				int e = errno;
