@@ -426,7 +426,7 @@ kbd_getch()
 			nchars = 1;
 #else /* !MSDOS */
 # ifdef WIN32
-			if (!charp()) {
+			if (UpdModLine || !charp()) {
 				redisplay();
 				flushscreen();
 			}
@@ -1201,6 +1201,8 @@ char	*argv[];
 				}
 				(void) div_wind(curwind, nwinds - 1);
 				break;
+			case '-':	/* Ignore -- which visudo provides */
+				break;
 			default:
 				error("Invalid switch %s",argv[1]);
 				/* NOTREACHED */
@@ -1326,14 +1328,14 @@ raw_complain(fmt, va_alist)
 	va_list	ap;
 	const char *bp;
 	size_t	rem;
-	SSIZE_T r;
+	JSSIZE_T r;
 
 	va_init(ap, fmt);
 	format(buf, sizeof(buf) - 2, fmt, ap);
 	va_end(ap);
 	strcat(buf, "\r\n");	/* \r *may* be redundant */
 	for (bp = buf, rem = strlen(buf)
-	; rem > 0 && (r = write(2, bp, rem)) != (SSIZE_T)rem
+	; rem > 0 && (r = write(2, bp, rem)) != (JSSIZE_T)rem
 	; bp += r, rem -=r) {
 		if (r < 0 || errno != EINTR)
 		    break;	/* give up */
